/*
 * Copyright (c) 2004-2007 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2006 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * Copyright (c) 2010      Cisco Systems, Inc. All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */
#include "ompi_config.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif  /* HAVE_UNISTD_H */
#include <stddef.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif  /* HAVE_STRING_H */
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif  /* HAVE_SYS_TYPES_H */
#include <sys/mman.h>
#include <dlfcn.h>

#include "ompi/constants.h"

#include "opal/class/opal_bitmap.h"
#include "opal/datatype/opal_convertor.h"
#include "opal/sys/atomic.h"

#include "opal/runtime/opal.h"
#include "opal/mca/event/event.h"
#include "opal/mca/base/mca_base_param.h"
#include "btl_hlink.h"
#include "btl_hlink_frag.h"

#ifdef BTL_HLINK_USE_CMEM
#include <ti/cmem.h>
#endif

#define SYNC_TOKEN   0xdeadbeef
#define SYNC_TIMEOUT 5
#define MAX_HOSTNAME_SIZE 80
//#ifdef BTL_HLINK_VERBOSE
#include <time.h>
/***************************************************************************************************/
static double clock_diff (struct timespec *start, struct timespec *end)
{
    return (double)(end->tv_sec - start->tv_sec) * 1000.0 + (double)(end->tv_nsec - start->tv_nsec) * 0.000001;
}
/***************************************************************************************************/
//#endif

#define MAP_SIZE 4096UL
#define MAP_MASK (MAP_SIZE - 1)

static int mca_btl_hlink_component_progress (void);
static int hlink_register(void);
/***********************************************************************/
static opal_event_t btl_hlink_term_handler1;
static opal_event_t btl_hlink_term_handler2;
static void btl_hlink_shutdown_callback(int fd, short flags, void *arg)
{
   opal_output(0, "BTL_HLINK_SHUTDOWN_CALLBACK!!!");
}
/***********************************************************************/
/*
 * Shared Memory (SELF) component instance. 
 */

mca_btl_hlink_component_t mca_btl_hlink_component = {
    {  /* super is being filled in */
        /* First, the mca_base_component_t struct containing meta information
          about the component itself */
        {
            MCA_BTL_BASE_VERSION_2_0_0,

            "hlink", /* MCA component name */
            OMPI_MAJOR_VERSION,  /* MCA component major version */
            OMPI_MINOR_VERSION,  /* MCA component minor version */
            OMPI_RELEASE_VERSION,  /* MCA component release version */
            mca_btl_hlink_component_open,  /* component open */
            mca_btl_hlink_component_close,  /* component close */
            NULL,
            hlink_register,
        },
        {
            /* The component is checkpoint ready */
            MCA_BASE_METADATA_PARAM_CHECKPOINT
        },

        mca_btl_hlink_component_init, 
        mca_btl_hlink_component_progress,
    }  /* end super */
};
/**************************************************************************************/
static inline char* mca_btl_hlink_param_register_string(
    const char* param_name,
    const char* default_value)
{
    char *param_value;
    int id = mca_base_param_register_string("btl","hlink",param_name,NULL,default_value);
    mca_base_param_lookup_string(id, &param_value);
    return param_value;
}

static inline int mca_btl_hlink_param_register_int(
    const char* param_name,
    int default_value)
{
    int id = mca_base_param_register_int("btl","hlink",param_name,NULL,default_value);
    int param_value = default_value;
    mca_base_param_lookup_int(id,&param_value);
    return param_value;
}
//TIMPIDBG
static int hl_fd = 0;
static char *map_base_hl0, *map_base_hl1;
static char *map_base_cfghl;
/*******/
static void timpi_dbg_hlink_start(void)
{
#ifdef BTL_HLINK_VERBOSE
  if((hl_fd = open("/dev/mem", O_RDWR | O_SYNC)) < 0) {
    printf ("\n[%c] TIMPI DBG CANNOT OPEN dev mem for debug!!!\n", mca_btl_hlink_component.node_id); fflush(stdout);
    return -1;
  }
  printf("/dev/mem opened.\n"); fflush(stdout);
  /* Map one page */
  map_base_hl0 = (char *)mmap(0, 2*MAP_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, hl_fd, 0x0231A000);
  map_base_hl1 = (char *)mmap(0, 2*MAP_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, hl_fd, 0x0231C000);
  map_base_cfghl = (char *)mmap(0, MAP_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, hl_fd, 0x21400000);
  *((unsigned int *)(map_base_cfghl +   0x4c)) = 0x0000; //reset ECC
  *((unsigned int *)(map_base_cfghl +  0x14c)) = 0x0000; //reset ECC

  usleep(500000);
#endif
}
static void timpi_dbg_hlink_end (void)
{
#ifdef BTL_HLINK_VERBOSE
  munmap(map_base_hl0, 2*MAP_SIZE);
  munmap(map_base_hl1, 2*MAP_SIZE);
  munmap(map_base_cfghl, MAP_SIZE);
  close(hl_fd);
  hl_fd = 0;
#endif
}
/*******/
static void timpi_dbg_serdes_dump(int chkpoint) {
unsigned int i, j, addr;
  if(hl_fd == 0) return;
  if(hl_fd < 0) {
    printf ("\n[%c] COULD NOT MAP dev mem for DBG!!\n", mca_btl_hlink_component.node_id); fflush(stdout);
    return;
  }
  for (i = 0; i < 64; i ++) 
  {
    addr = 0x0000 + i * 4;
    printf ("\n[%c] %03d CMU0 SERDES(%04x) hl0:%08x hl1:%08x)",
             mca_btl_hlink_component.node_id, chkpoint, addr,
             *((unsigned int *)(map_base_hl0 + 0x0000 + i * 4)),
             *((unsigned int *)(map_base_hl1 + 0x0000 + i * 4)));
  }
  for (j = 0; j < 4; j ++) 
  {
    for (i = 0; i < 128; i ++) 
    {
      addr = 0x200 + 0x200 * j + i * 4;
      printf ("\n[%c] %03d LANE%d SERDES(%04x) hl0:%08x hl1:%08x)",
         mca_btl_hlink_component.node_id, chkpoint, j, addr,
        *((unsigned int *)(map_base_hl0 + addr)),
        *((unsigned int *)(map_base_hl1 + addr)));
    }
  }
  for (i = 0; i < 128; i ++) 
  {
    addr = 0x0A00 + i * 4;
    printf ("\n[%c] %03d COMLANE SERDES(%04x) hl0:%08x hl1:%08x)",
             mca_btl_hlink_component.node_id, chkpoint, addr,
             *((unsigned int *)(map_base_hl0 + addr)),
             *((unsigned int *)(map_base_hl1 + addr)));
  }
  for (i = 0; i < 16; i ++) 
  {
    addr = 0x1FC0 + i * 4;
    printf ("\n[%c] %03d WIZ SERDES(%04x) hl0:%08x hl1:%08x)",
             mca_btl_hlink_component.node_id, chkpoint, addr,
             *((unsigned int *)(map_base_hl0 + addr)),
             *((unsigned int *)(map_base_hl1 + addr)));
  }
}
/*******/
void timpi_dbg_hlink(int chkpoint) {
if(hl_fd == 0) return;
if(hl_fd < 0) {
  return;
}
printf ("\n[%c] %03d TIMPIDBG HlnkSERDES(0x1ff4) hl0:%08x hl1:%08x)",
    mca_btl_hlink_component.node_id, chkpoint, *((unsigned int *)(map_base_hl0 + 0x1ff4)), *((unsigned int *)(map_base_hl1 + 0x1ff4)));
printf ("\n[%c] %03d TIMPIDBG HlnkSERDL0(0x1fe0) hl0:%08x hl1:%08x)",
    mca_btl_hlink_component.node_id, chkpoint, *((unsigned int *)(map_base_hl0 + 0x1fe0)), *((unsigned int *)(map_base_hl1 + 0x1fe0)));
printf ("\n[%c] %03d TIMPIDBG HlnkSERDL1(0x1fe4) hl0:%08x hl1:%08x)",
    mca_btl_hlink_component.node_id, chkpoint, *((unsigned int *)(map_base_hl0 + 0x1fe4)), *((unsigned int *)(map_base_hl1 + 0x1fe4)));
printf ("\n[%c] %03d TIMPIDBG HlnkSERDL2(0x1fe8) hl0:%08x hl1:%08x)",
    mca_btl_hlink_component.node_id, chkpoint, *((unsigned int *)(map_base_hl0 + 0x1fe8)), *((unsigned int *)(map_base_hl1 + 0x1fe8)));
printf ("\n[%c] %03d TIMPIDBG HlnkSERDL3(0x1fec) hl0:%08x hl1:%08x)",
    mca_btl_hlink_component.node_id, chkpoint, *((unsigned int *)(map_base_hl0 + 0x1fec)), *((unsigned int *)(map_base_hl1 + 0x1fec)));

printf ("\n[%c] %03d TIMPIDBG HyperLinkCFG(0x00) hl0:%08x hl1:%08x)",
    mca_btl_hlink_component.node_id, chkpoint, *((unsigned int *)(map_base_cfghl + 0x0)), *((unsigned int *)(map_base_cfghl + 0x100))); 
printf ("\n[%c] %03d TIMPIDBG HyperLinkCFG(0x04) hl0:%08x hl1:%08x)",
    mca_btl_hlink_component.node_id, chkpoint, *((unsigned int *)(map_base_cfghl + 0x4)), *((unsigned int *)(map_base_cfghl + 0x104))); 
printf ("\n[%c] %03d TIMPIDBG HyperLinkCFG(0x08) hl0:%08x hl1:%08x)",
    mca_btl_hlink_component.node_id, chkpoint, *((unsigned int *)(map_base_cfghl + 0x8)), *((unsigned int *)(map_base_cfghl + 0x108))); 
printf ("\n[%c] %03d TIMPIDBG HyperLinkCFG(0x44) hl0:%08x hl1:%08x)",
    mca_btl_hlink_component.node_id, chkpoint, *((unsigned int *)(map_base_cfghl + 0x44)), *((unsigned int *)(map_base_cfghl + 0x144))); 
printf ("\n[%c] %03d TIMPIDBG HyperLinkCFG(0x4c) hl0:%08x hl1:%08x)",
    mca_btl_hlink_component.node_id, chkpoint, *((unsigned int *)(map_base_cfghl + 0x4c)), *((unsigned int *)(map_base_cfghl + 0x14c))); 
printf ("\n[%c] %03d TIMPIDBG HyperLinkCFG(0x58) hl0:%08x hl1:%08x)\n",
    mca_btl_hlink_component.node_id, chkpoint, *((unsigned int *)(map_base_cfghl + 0x58)), *((unsigned int *)(map_base_cfghl + 0x158))); 
fflush(stdout); 
}
/*******/
int timpi_dbg_hlink_health(int chkpoint) {
int i, err_flag = 0;
unsigned int ecchl0, ecchl1;
unsigned int serdeshl0, serdeshl1;
unsigned int serdeslanehl0[4];
unsigned int serdeslanehl1[4];
  if(hl_fd == 0) return 0;
  if(hl_fd < 0) {
    return 1;
  }
  ecchl0    =  *((unsigned int *)(map_base_cfghl +   0x4c));
  ecchl1    =  *((unsigned int *)(map_base_cfghl +  0x14c));
  serdeshl0 =  *((unsigned int *)(map_base_hl0   + 0x1ff4));
  serdeshl1 =  *((unsigned int *)(map_base_hl1   + 0x1ff4));
  serdeslanehl0[0] = *((unsigned int *)(map_base_hl1   + 0x1fe0));
  serdeslanehl0[1] = *((unsigned int *)(map_base_hl1   + 0x1fe4)); 
  serdeslanehl0[2] = *((unsigned int *)(map_base_hl1   + 0x1fe8));
  serdeslanehl0[3] = *((unsigned int *)(map_base_hl1   + 0x1fec));
  serdeslanehl1[0] = *((unsigned int *)(map_base_hl1   + 0x1fe0));
  serdeslanehl1[1] = *((unsigned int *)(map_base_hl1   + 0x1fe4));
  serdeslanehl1[2] = *((unsigned int *)(map_base_hl1   + 0x1fe8));
  serdeslanehl1[3] = *((unsigned int *)(map_base_hl1   + 0x1fec));

  if(ecchl0 || ecchl1) 
  {
    printf ("\n[%c] %03d TIMPIDBG <HyperLinkCFG_ERROR>  hl0:%08x hl1:%08x)", mca_btl_hlink_component.node_id, chkpoint, ecchl0, ecchl1);
    goto HL_ERROR_RPT;
  }
  if(((serdeshl0 & 0x01) == 0) || ((serdeshl1 & 0x01) == 0)) 
  {
    printf ("\n[%c] %03d TIMPIDBG <HyperLinkSERDES_ERROR>  hl0:%08x hl1:%08x)", mca_btl_hlink_component.node_id, chkpoint, serdeshl0, serdeshl1);
    goto HL_ERROR_RPT;
  }
  for (i = 0; i < 4; i ++) {
    if((serdeslanehl0[i] & 0x03) != 2) {
      printf ("\n[%c] %03d TIMPIDBG <HyperLinkSERDES_LANELOSS_ERROR>  Lane%d hl0:%08x)", mca_btl_hlink_component.node_id, chkpoint, i, serdeslanehl0[i]);
      goto HL_ERROR_RPT;
    }
    if((serdeslanehl1[i] & 0x03) != 2) {
      printf ("\n[%c] %03d TIMPIDBG <HyperLinkSERDES_LANELOSS_ERROR>  Lane%d hl1:%08x)", mca_btl_hlink_component.node_id, chkpoint, i, serdeslanehl1[i]);
      goto HL_ERROR_RPT;
    }
  }
  return 0;
/***ERROR detected!!!***/
HL_ERROR_RPT:
    timpi_dbg_hlink(chkpoint);
    fflush(stdout);
    return 1;
}
/*******/
#if 0
static void hlink_sync_nodes (int timeout)
{
 char *rmt_hl_hostname = (char *)mca_btl_hlink_component.hl_win_rmt_hostname;
 char *rmt1_hl_hostname = (char *)mca_btl_hlink_component.hl_win_rmt1_hostname;
 char *rmt_bridge_hl_hostname = (char *)mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname;
 struct timespec tp_start, tp_end;
 unsigned int sync_word_hl0       = 0xffffffff;
 unsigned int sync_word_hl1       = 0xffffffff;
 unsigned int sync_word_hl_bridge = 0xffffffff;
 int hl0_detected = 0, hl1_detected = 0, hl_bridge_detected = 0;
 int ii, jj = -1;
   opal_output_verbose(0, mca_btl_base_output, "WAIT FOR SYNC!"); 
   clock_gettime(CLOCK_MONOTONIC, &tp_start);

   for (ii = 0; ii < timeout; ii ++)
   {
      hl0_detected = hl1_detected = 0;
      if(mca_btl_hlink_component.use_hl0)
      {
        sync_word_hl0 = *((unsigned int *)(rmt_hl_hostname  + MAX_HOSTNAME_SIZE - sizeof(unsigned int)));
        if(sync_word_hl0 == SYNC_TOKEN) {
          hl0_detected = 1;
          if(!mca_btl_hlink_component.use_hl1) break;
        }
      }
      if(mca_btl_hlink_component.use_hl1)
      {
        sync_word_hl1 = *((unsigned int *)(rmt1_hl_hostname + MAX_HOSTNAME_SIZE - sizeof(unsigned int)));
        if(sync_word_hl1 == SYNC_TOKEN) {
          hl1_detected = 1;
          if(!mca_btl_hlink_component.use_hl1) break;
        }
      } 
      if(hl0_detected * hl1_detected) {
         break;
      } 
      sleep(1);
   }
   mca_btl_hlink_component.use_hl0 = hl0_detected;
   mca_btl_hlink_component.use_hl1 = hl1_detected;
   clock_gettime(CLOCK_MONOTONIC, &tp_end);
   opal_output_verbose(0, mca_btl_base_output, 
       "%d(%d) time %lfms hostnames REMOTE TOKEN HL0 %08x (%c%c%c%c) HL1 %08x (%c%c%c%c) BRIDGE %08x (%c%c%c%c)\n",
       ii, jj, clock_diff(&tp_start, &tp_end),
       sync_word_hl0, 
       rmt_hl_hostname[0], rmt_hl_hostname[1], rmt_hl_hostname[2], rmt_hl_hostname[3],
       sync_word_hl1, 
       rmt1_hl_hostname[0], rmt1_hl_hostname[1], rmt1_hl_hostname[2], rmt1_hl_hostname[3],
       sync_word_hl_bridge,
       rmt_bridge_hl_hostname[0], rmt_bridge_hl_hostname[1], rmt_bridge_hl_hostname[2], rmt_bridge_hl_hostname[3]
   );
}
#endif
/*******/
static int hlink_register(void)
{
    int i;

    opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK TIMPIDBG: hlink_component_register!!!");
    /**************************************/
    mca_btl_hlink_component.node_id = '1';
    { /** Find location in topology in device-tree: if this is EVM, default to 1 **/
      FILE *node_id_detect = fopen("/proc/device-tree/sl1500/node/id", "rb");
      unsigned char node_id_buff[2];
      if(node_id_detect)
      {
         fread(node_id_buff, 1, 2, node_id_detect);
         mca_btl_hlink_component.node_id = node_id_buff[0];
         fclose (node_id_detect);
      } else {
        opal_output_verbose(100, mca_btl_base_output, "This is EVM, using hl0 only!");
      }
    }

    //TIMPIDBG
    timpi_dbg_hlink_start();
    /* register HLINK component parameters */
    if(mca_btl_hlink_component.node_id == '1')
    {
      mca_btl_hlink_component.ddr_loc_addr =
        mca_btl_hlink_param_register_int("ddr_loc_addr", BTL_HLINK_DDR_LOC_ADDRESS);
      mca_btl_hlink_component.msmc_loc_addr =
        mca_btl_hlink_param_register_int("msmc_loc_addr", BTL_HLINK_MSMC_LOC_ADDRESS);
      mca_btl_hlink_component.ddr_rmt_addr =
        mca_btl_hlink_param_register_int("ddr_rmt_addr", BTL_HLINK_DDR_RMT_ADDRESS);
      mca_btl_hlink_component.msmc_rmt_addr =
        mca_btl_hlink_param_register_int("msmc_rmt_addr", BTL_HLINK_MSMC_RMT_ADDRESS);

      mca_btl_hlink_component.ddr_loc1_addr =
        mca_btl_hlink_param_register_int("ddr_loc1_addr", BTL_HLINK_DDR_LOC1_ADDRESS);
      mca_btl_hlink_component.msmc_loc1_addr =
        mca_btl_hlink_param_register_int("msmc_loc1_addr", BTL_HLINK_MSMC_LOC1_ADDRESS);
      mca_btl_hlink_component.ddr_rmt1_addr =
        mca_btl_hlink_param_register_int("ddr_rmt1_addr", BTL_HLINK_DDR_RMT1_ADDRESS);
      mca_btl_hlink_component.msmc_rmt1_addr =
        mca_btl_hlink_param_register_int("msmc_rmt1_addr", BTL_HLINK_MSMC_RMT1_ADDRESS);
    } else { /** Mirrored addresses **/
      mca_btl_hlink_component.ddr_loc1_addr =
        mca_btl_hlink_param_register_int("ddr_loc_addr", BTL_HLINK_DDR_LOC_ADDRESS);
      mca_btl_hlink_component.msmc_loc1_addr =
        mca_btl_hlink_param_register_int("msmc_loc_addr", BTL_HLINK_MSMC_LOC_ADDRESS);
      mca_btl_hlink_component.ddr_rmt1_addr =
        mca_btl_hlink_param_register_int("ddr_rmt_addr", BTL_HLINK_DDR_RMT_ADDRESS);
      mca_btl_hlink_component.msmc_rmt1_addr =
        mca_btl_hlink_param_register_int("msmc_rmt_addr", BTL_HLINK_MSMC_RMT_ADDRESS);

      mca_btl_hlink_component.ddr_loc_addr =
        mca_btl_hlink_param_register_int("ddr_loc1_addr", BTL_HLINK_DDR_LOC1_ADDRESS);
      mca_btl_hlink_component.msmc_loc_addr =
        mca_btl_hlink_param_register_int("msmc_loc1_addr", BTL_HLINK_MSMC_LOC1_ADDRESS);
      mca_btl_hlink_component.ddr_rmt_addr =
        mca_btl_hlink_param_register_int("ddr_rmt1_addr", BTL_HLINK_DDR_RMT1_ADDRESS);
      mca_btl_hlink_component.msmc_rmt_addr =
        mca_btl_hlink_param_register_int("msmc_rmt1_addr", BTL_HLINK_MSMC_RMT1_ADDRESS);
    }

    mca_btl_hlink_component.msmc_rmt_bridge_addr =
        mca_btl_hlink_param_register_int("msmc_rmt_bridge_addr", BTL_HLINK_MSMC_BRIDGE_ADDRESS);
    mca_btl_hlink_component.ddr_rmt_bridge_addr =
        mca_btl_hlink_param_register_int("ddr_rmt_bridge_addr", BTL_HLINK_DDR_BRIDGE_ADDRESS);

    mca_btl_hlink_component.fifo_size =
        mca_btl_hlink_param_register_int("fifo_size", BTL_HLINK_MAX_FIFO_ITEMS);
    mca_btl_hlink_component.use_hl0  = 
        mca_btl_hlink_param_register_int("use_hl0", 1);
    mca_btl_hlink_component.use_hl1  = 
        mca_btl_hlink_param_register_int("use_hl1", 1);
    mca_btl_hlink_component.serdes_init  = 
        mca_btl_hlink_param_register_int("serdes_init", 0);
    mca_btl_hlink_component.use_edma  =
        mca_btl_hlink_param_register_int("use_edma", BTL_HLINK_USE_EDMA);


    if (access("/dev/edma3", F_OK)) { 
      mca_btl_hlink_component.use_edma = 0; 
    }

    mca_base_param_reg_int( (mca_base_component_t*)&mca_btl_hlink_component, "free_list_num",
                            "Number of fragments by default", false, false,
                            0, &mca_btl_hlink_component.free_list_num );
    mca_base_param_reg_int( (mca_base_component_t*)&mca_btl_hlink_component, "free_list_max",
                            "Maximum number of fragments", false, false,
                            -1, &mca_btl_hlink_component.free_list_max );
    mca_base_param_reg_int( (mca_base_component_t*)&mca_btl_hlink_component, "free_list_inc",
                            "Increment by this number of fragments", false, false,
                            32, &mca_btl_hlink_component.free_list_inc ); //DJDBG, used to be 32

      return OMPI_SUCCESS;
}
/*
 *  Called by MCA framework to open the component, registers
 *  component parameters.
 */
int mca_btl_hlink_component_open(void)
{
    opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG: hlink_component_open!!!");
    mca_btl_hlink.super.btl_exclusivity = MCA_BTL_EXCLUSIVITY_HIGH;
    mca_btl_hlink.super.btl_eager_limit = BTL_HLINK_EAGER_SIZE;
    mca_btl_hlink.super.btl_rndv_eager_limit = BTL_HLINK_EAGER_SIZE;
    mca_btl_hlink.super.btl_max_send_size = BTL_HLINK_MAX_SEND_SIZE;
    mca_btl_hlink.super.btl_flags = MCA_BTL_FLAGS_SEND | MCA_BTL_FLAGS_SEND_INPLACE;
    mca_btl_hlink.super.btl_seg_size = sizeof (mca_btl_base_segment_t);
    mca_btl_hlink.super.btl_bandwidth = 40000;
    mca_btl_hlink.super.btl_latency   = 1;
    mca_btl_base_param_register(&mca_btl_hlink_component.super.btl_version, &mca_btl_hlink);
    /**************************************/
    mca_btl_hlink_component.max_frag_size = mca_btl_hlink.super.btl_max_send_size;
    mca_btl_hlink_component.eager_limit = mca_btl_hlink.super.btl_eager_limit;
    /**************************************/
    /* initialize objects */
    OBJ_CONSTRUCT(&mca_btl_hlink_component.hlink_lock, opal_mutex_t);
    OBJ_CONSTRUCT(&mca_btl_hlink_component.hlink_frags_eager, ompi_free_list_t);
    OBJ_CONSTRUCT(&mca_btl_hlink_component.hlink_frags_max, ompi_free_list_t);
    {
       extern void *hyplnkInitCfg[];
       int ii;
       opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK  BTL HLINK start of HYPLNKINITCFG: %p ", hyplnkInitCfg);
       if(hyplnkInitCfg) {
         for (ii = 0; ii < 8; ii ++) opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK [%p]", hyplnkInitCfg[ii]);
       }
       opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK BTL HLINK end of HYPLNKINITCFG\n");     
    }
    /**************************************/
    mca_btl_hlink_component.hl_win_edma_transfer = NULL;
#ifdef BTL_HLINK_USE_CMEM
    if(CMEM_init() != 0) {
      opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK: CMEM_init not available!");
    } else {
      opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK: CMEM_init OK!");
    }
#endif
    return OMPI_SUCCESS;
}

/*
 * component cleanup - sanity checking of queue lengths
 */

int mca_btl_hlink_component_close(void)
{
    opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG: hlink_component_close!!!");
    OBJ_DESTRUCT(&mca_btl_hlink_component.hlink_lock);
    OBJ_DESTRUCT(&mca_btl_hlink_component.hlink_frags_eager);
    OBJ_DESTRUCT(&mca_btl_hlink_component.hlink_frags_max);
    return OMPI_SUCCESS;
}

/*
 *  HLINK component initialization
 */
mca_btl_base_module_t** mca_btl_hlink_component_init( int *num_btls, 
                                                     bool enable_progress_threads,
                                                     bool enable_mpi_threads )
{
    mca_btl_base_module_t **btls = NULL;
    char host_name[MAX_HOSTNAME_SIZE], *ptr_hostname;
    char stage_hostname[MAX_HOSTNAME_SIZE];
    int rc;
    struct timespec tp_start, tp_end;
    /** mpm_transport init **/
    mpm_transport_open_t ocfg = {
      .open_mode      = (O_SYNC|O_RDWR),
      .msec_timeout   = BTL_HLINK_SYNC_TIMEOUT,
      .serdes_init    = 0
    };

    mpm_transport_mmap_t mcfg = {
      .mmap_prot      = (PROT_READ|PROT_WRITE),
      .mmap_flags     = MAP_SHARED,
    };

    gethostname(host_name, MAX_HOSTNAME_SIZE);
    opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG: hlink_component_init!!!");

    /*** Enable and initialize peripheral ***/
#if 0
{
    mpm_transport_open_t ocfg = {
      .open_mode      = (O_SYNC|O_RDWR),
      .msec_timeout   = BTL_HLINK_SYNC_TIMEOUT,
      .serdes_init    = 0
    };

    if(mpm_transport_peripheral_enable("arm-remote-hyplnk-0", &ocfg)) {
       opal_output_verbose(0, mca_btl_base_output, "BTL_HLINK TIMPIDBG: hyplnk peripheral enable FAILED!!!");
       return OMPI_ERROR;
    }
    //TODO: EDMA3 initialization required!
    opal_output_verbose(0, mca_btl_base_output, "BTL_HLINK TIMPIDBG: hyplnk peripheral enable OK!!!");
    sleep(3);
}
    /*** Enable and initialize peripheral ***/
#else
    if(mca_btl_hlink_component.serdes_init) {
      system("/usr/sbin/mpmcl transport arm-remote-hyplnk-0 open");
      sleep(3); //Above command just sends non-blocking message, leave some time for nodes to connect
    }
#endif

    /* allocate the Shared Memory PTL */
    *num_btls = 1;
    btls = (mca_btl_base_module_t**)malloc((*num_btls)*sizeof(mca_btl_base_module_t*));
    if (NULL == btls) {
        return NULL;
    }
    mca_btl_hlink_component.num_hlink_btls = 1;
    mca_btl_hlink_component.hlink_btls     = btls;
    mca_btl_hlink_component.mpm_transport_hyplnk0 = NULL;
    mca_btl_hlink_component.mpm_transport_hyplnk1 = NULL;
    /* initialize free lists */
    rc = ompi_free_list_init_ex_new (&mca_btl_hlink_component.hlink_frags_eager,
                                     sizeof(mca_btl_hlink_frag1_t),
                                     opal_cache_line_size, OBJ_CLASS(mca_btl_hlink_frag1_t),
                                     mca_btl_hlink.super.btl_eager_limit,
                                     opal_cache_line_size,
                                     mca_btl_hlink_component.free_list_num,
                                     mca_btl_hlink_component.free_list_max,
                                     mca_btl_hlink_component.free_list_inc,
                                     NULL, (ompi_free_list_item_init_fn_t) mca_btl_hlink_frag_init,
                                     (void *) &mca_btl_hlink);

    if (OPAL_UNLIKELY(OMPI_SUCCESS != rc)) {
        opal_output(0, "error creating hlink fragment1 free list");
        return rc;
    }


    rc = ompi_free_list_init_ex_new (&mca_btl_hlink_component.hlink_frags_max,
                                     sizeof(mca_btl_hlink_frag1_t),
                                     opal_cache_line_size, OBJ_CLASS(mca_btl_hlink_frag1_t),
                                     mca_btl_hlink.super.btl_max_send_size,
                                     opal_cache_line_size,
                                     mca_btl_hlink_component.free_list_num,
                                     mca_btl_hlink_component.free_list_max,
                                     mca_btl_hlink_component.free_list_inc,
                                     NULL, (ompi_free_list_item_init_fn_t) mca_btl_hlink_frag_init,
                                     (void *) &mca_btl_hlink);

    if (OPAL_UNLIKELY(OMPI_SUCCESS != rc)) {
        opal_output(0, "error creating hlink fragment2 free list");
        return rc;
    }
    /* get pointer to the btls */
    btls[0] = (mca_btl_base_module_t *)(&mca_btl_hlink);
      /************************************************************************************************************************/
      /** SHARED MEMORY ACCESS: one MSMC region is mapped for both interfaces, and two DDR regions, one per HYPLNK interface **/
      /************************************************************************************************************************/
      /** Open local shared memory interface **/
      mca_btl_hlink_component.mpm_transport_shmem = (void *)mpm_transport_open("arm-local-shmem", &ocfg);
      if (!mca_btl_hlink_component.mpm_transport_shmem) {
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem open failed");
        return NULL;
      }
      opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem open successfull!!");
      mca_btl_hlink_component.dma_send = NULL;
      mca_btl_hlink_component.dma_pingpong = 0;
#ifdef BTL_HLINK_USE_CMEM
      {
        CMEM_AllocParams  alloc_paramsHeap;
        uint64_t ph_addr;
        int32_t num_Blocks;

        CMEM_getNumBlocks(&num_Blocks);
        if (num_Blocks < 3)
        {
          opal_output(0, "MPI is using last CMEM block. At least 3 are expected and only %d are detected.\nBTL_SRIO won't be used.", num_Blocks);
          return NULL;
        }

        alloc_paramsHeap.type = CMEM_HEAP;
        alloc_paramsHeap.alignment = 4;
#ifdef BTL_HLINK_USE_CACHED
        alloc_paramsHeap.flags = CMEM_CACHED;
#else
        alloc_paramsHeap.flags = CMEM_NONCACHED;
#endif
        mca_btl_hlink_component.hl_win_edma_transfer = 
          CMEM_alloc2( num_Blocks - 1, BTL_HLINK_EDMA_MMAP_SIZE + BTL_HLINK_MMAP_SIZE * 3, &alloc_paramsHeap);
        if(mca_btl_hlink_component.hl_win_edma_transfer) {
          ph_addr = CMEM_getPhys (mca_btl_hlink_component.hl_win_edma_transfer);
          mca_btl_hlink_component.scratch_edma_transfer = (uint32_t)(ph_addr + 0x080000000LL);
          opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK: CMEM physAddr:%16llx (to %08lx) userAddr:%p", ph_addr,
            mca_btl_hlink_component.scratch_edma_transfer, mca_btl_hlink_component.hl_win_edma_transfer);
        } else {
          opal_output(0, "BTL_HLINK: ERROR: CMEM_alloc2(). Not using HLINK!");
          CMEM_exit();
          return NULL;
        }
      }
#else
      if(mca_btl_hlink_component.use_edma)
      {
        /** Map local MSMC region to user space, used for EDMA driven transfers **/
        mca_btl_hlink_component.hl_win_edma_transfer = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_shmem,
                                                                          mca_btl_hlink_component.scratch_edma_transfer, BTL_HLINK_EDMA_MMAP_SIZE, &mcfg);

        if(!mca_btl_hlink_component.hl_win_edma_transfer)
        {
          opal_output(0, "mmap_failed_hl_win_edma_transfer!");
          return NULL;
        }
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem MSMC of EDMA transfer mmap successfull (%p)!!",
          mca_btl_hlink_component.hl_win_edma_transfer);
      }
#endif
      /****-----------------------------------------------------****/
      /** Map local msmc region to user space, used over hyplnk0 **/
      /****-----------------------------------------------------****/
#ifdef BTL_HLINK_COALESCE_MSMC_WINDOWS
      mca_btl_hlink_component.hl_win_msmc_loc = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_shmem, 
                                                                    mca_btl_hlink_component.msmc_loc_addr, 2 * BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
#else
      mca_btl_hlink_component.hl_win_msmc_loc = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_shmem, 
                                                                    mca_btl_hlink_component.msmc_loc_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
#endif
      if(!mca_btl_hlink_component.hl_win_msmc_loc)
      {
        opal_output_verbose(0, mca_btl_base_output, "BTL_HLINK mmap_failed_hl_win_msmc_loc (local MSMC over hyplnk0)!");
        return NULL;
      }
memset(mca_btl_hlink_component.hl_win_msmc_loc, 0, BTL_HLINK_MSMC_MMAP_SIZE);
btl_hlink_atomic_smb();
      /* Identify yourself to remote endpoint, over hyplnk0 - needed for add_procs */
      ptr_hostname = (char *)mca_btl_hlink_component.hl_win_msmc_loc + BTL_HLINK_MSG_OFFSET - MAX_HOSTNAME_SIZE;
      memset (stage_hostname, 0, MAX_HOSTNAME_SIZE);
      strcpy (stage_hostname, host_name);
      *((unsigned int *)(stage_hostname + MAX_HOSTNAME_SIZE - sizeof(unsigned int))) = SYNC_TOKEN;
      btl_hlink_atomic_smb();
      memcpy (ptr_hostname, stage_hostname, MAX_HOSTNAME_SIZE);
      btl_hlink_atomic_smb();
      opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem MSMC0 mmap successfull!!");
#ifdef BTL_HLINK_USE_CMEM
      if(mca_btl_hlink_component.node_id == '1')
      {
        mca_btl_hlink_component.ddr_loc_addr = mca_btl_hlink_component.scratch_edma_transfer + BTL_HLINK_EDMA_MMAP_SIZE;
        mca_btl_hlink_component.hl_win_loc   = mca_btl_hlink_component.hl_win_edma_transfer  + BTL_HLINK_EDMA_MMAP_SIZE;
      } else {
        mca_btl_hlink_component.ddr_loc_addr =
          mca_btl_hlink_component.scratch_edma_transfer + BTL_HLINK_EDMA_MMAP_SIZE + BTL_HLINK_MMAP_SIZE;
        mca_btl_hlink_component.hl_win_loc   = 
          mca_btl_hlink_component.hl_win_edma_transfer  + BTL_HLINK_EDMA_MMAP_SIZE + BTL_HLINK_MMAP_SIZE;
      }
      mca_btl_hlink_component.ddr_rmt_addr =  mca_btl_hlink_component.ddr_loc_addr;
#else
      /** Map local DDR region to user space, used over hyplnk0 **/
      mca_btl_hlink_component.hl_win_loc = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_shmem,
                                                               mca_btl_hlink_component.ddr_loc_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
      if(!mca_btl_hlink_component.hl_win_loc)
      {
        opal_output (0, "mmap_failed_hl_win_loc (local DDR over hyplnk0)!");
        return NULL;
      }
      opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem DDR0 mmap successfull!!");
#endif
      /** Map local msmc region to user space, used over hyplnk1 **/
#ifdef BTL_HLINK_COALESCE_MSMC_WINDOWS
      mca_btl_hlink_component.hl_win_msmc_loc1 = mca_btl_hlink_component.hl_win_msmc_loc + BTL_HLINK_MSMC_MMAP_SIZE; 
#else
      mca_btl_hlink_component.hl_win_msmc_loc1 = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_shmem,
                                                                    mca_btl_hlink_component.msmc_loc1_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
#endif
      if(!mca_btl_hlink_component.hl_win_msmc_loc1)
      {
        opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK mmap_failed_hl_win_msmc_loc (local MSMC, over hyplnk1)!");
        return NULL;
      }
memset(mca_btl_hlink_component.hl_win_msmc_loc1, 0, BTL_HLINK_MSMC_MMAP_SIZE);
btl_hlink_atomic_smb();
      /* Identify yourself to remote endpoint, over hyplnk0 - needed for add_procs */
      ptr_hostname = (char *)mca_btl_hlink_component.hl_win_msmc_loc1 + BTL_HLINK_MSG_OFFSET - MAX_HOSTNAME_SIZE;
      memset (stage_hostname, 0, MAX_HOSTNAME_SIZE);
      strcpy (stage_hostname, host_name);
      *((unsigned int *)(stage_hostname + MAX_HOSTNAME_SIZE - sizeof(unsigned int))) = SYNC_TOKEN;
      btl_hlink_atomic_smb();
      memcpy (ptr_hostname, stage_hostname, MAX_HOSTNAME_SIZE);
      btl_hlink_atomic_smb();
      opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem MSMC0 mmap successfull!!");
#ifdef BTL_HLINK_USE_CMEM
      if(mca_btl_hlink_component.node_id == '1')
      {
        mca_btl_hlink_component.ddr_loc1_addr = 
          mca_btl_hlink_component.scratch_edma_transfer + BTL_HLINK_EDMA_MMAP_SIZE + BTL_HLINK_MMAP_SIZE;
        mca_btl_hlink_component.hl_win_loc1   = 
          mca_btl_hlink_component.hl_win_edma_transfer    + BTL_HLINK_EDMA_MMAP_SIZE + BTL_HLINK_MMAP_SIZE;
      } else {
        mca_btl_hlink_component.ddr_loc1_addr =
          mca_btl_hlink_component.scratch_edma_transfer + BTL_HLINK_EDMA_MMAP_SIZE;
        mca_btl_hlink_component.hl_win_loc1   =
          mca_btl_hlink_component.hl_win_edma_transfer    + BTL_HLINK_EDMA_MMAP_SIZE;
      }
      mca_btl_hlink_component.ddr_rmt1_addr = mca_btl_hlink_component.ddr_loc1_addr;
      mca_btl_hlink_component.ddr_rmt_bridge_addr = 
        mca_btl_hlink_component.scratch_edma_transfer + BTL_HLINK_EDMA_MMAP_SIZE + 2 * BTL_HLINK_MMAP_SIZE;
#else
      /** Map local DDR region to user space, used over hyplnk1 **/
      mca_btl_hlink_component.hl_win_loc1 = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_shmem,
                                                                    mca_btl_hlink_component.ddr_loc1_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
      if(!mca_btl_hlink_component.hl_win_loc1)
      {
        opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK mmap_failed_hl_win_loc (local DDR, over hyplnk1)!");
        return NULL;
      }
      opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem DDR1 mmap successfull!!");
#endif
   /*********************************************************************************************/
   /** Probe and Open hyplnk0 interface                                                        **/
   /*********************************************************************************************/
  //mpm_transport_set_verbosity(mca_btl_base_output / 10);

/***
base=0x21400000 for hyperlink 1 or 0x21400100 for hyperlink 2:
what is in lane power management control register (base+0x44)?  See 3.2.15 in UG. 
I.d try seeting the low 3 bits to 010b and 100b to see what happens.  
This disables dynamic lane enabling and forces 1 (010b) or 4 (100b) lanes. 

the ECC error counters at base + 0x4c should stay 0.  
If it increases, there are bit errors.  If they happen at a noticeable rate such that each read shows an increase, 
there is meaningful probability of double bit error. If it gets a double bit error, locks up.

link status register at base + 0x58 is interesting.  
If the lane power management control register has 111b in low bits, then it will dynamically change number of lanes.  
This register tells you how many are currently active.  I don.t trust the dynamic behavior, especially with snow bush.
***/
   timpi_dbg_hlink(0);

   if(mca_btl_hlink_component.node_id == '1') 
   { /* This block is also executed by EVMs */
      opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK attempt HyperLink0 then HyperLink1");
      if(mca_btl_hlink_component.use_hl0)
      { /* If user requested not to use Hyperlink0, by resetting use_hl0 to 0, this is skipped */
        /** Open hyplnk0 interface **/
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 attempt opening");
        if(mca_btl_hlink_component.use_edma) {
          mca_btl_hlink_component.mpm_transport_hyplnk0 = (void *)mpm_transport_open("arm-remote-hyplnk-0", &ocfg);
        } else {
          mca_btl_hlink_component.mpm_transport_hyplnk0 = (void *)mpm_transport_open("arm-remote-hyplnk-0-no-dma", &ocfg);
        }
        if (!mca_btl_hlink_component.mpm_transport_hyplnk0) {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 open failed");
          mca_btl_hlink_component.use_hl0 = 0;
        } else {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 open successfull!!");
        }
      }
      if(mca_btl_hlink_component.use_hl1)
      { /* If user requested not to use Hyperlink1, by resetting use_hl1 to 0, this is skipped */
        /** Open hyplnk1 interface **/
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 attempt opening");
        if(mca_btl_hlink_component.use_edma) {
          mca_btl_hlink_component.mpm_transport_hyplnk1 = (void *)mpm_transport_open("arm-remote-hyplnk-1", &ocfg);
        } else {
          mca_btl_hlink_component.mpm_transport_hyplnk1 = (void *)mpm_transport_open("arm-remote-hyplnk-1-no-dma", &ocfg);
        }
        if (!mca_btl_hlink_component.mpm_transport_hyplnk1) {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 open failed");
          mca_btl_hlink_component.use_hl1 = 0;
        } else {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 open successfull!!");
        }
      }
   } else { /* other nodes in cluster; this code is not used for EVM */
      opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK attempt HyperLink1 then HyperLink0");
      if(mca_btl_hlink_component.use_hl1)
      { /* If user requested not to use Hyperlink1, by resetting use_hl1 to 0, this is skipped */
        /** Open hyplnk1 interface **/
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 attempt opening");
        if(mca_btl_hlink_component.use_edma) {
          mca_btl_hlink_component.mpm_transport_hyplnk1 = (void *)mpm_transport_open("arm-remote-hyplnk-1", &ocfg);
        } else {
          mca_btl_hlink_component.mpm_transport_hyplnk1 = (void *)mpm_transport_open("arm-remote-hyplnk-1-no-dma", &ocfg);
        }
        if (!mca_btl_hlink_component.mpm_transport_hyplnk1) {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 open failed");
          mca_btl_hlink_component.use_hl1 = 0;
        } else {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 open successfull!!");
        }
      }
      if(mca_btl_hlink_component.use_hl0)
      { /* If user requested not to use Hyperlink0, by resetting use_hl0 to 0, this is skipped */
        /** Open hyplnk0 interface **/
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 attempt opening\n");
        if(mca_btl_hlink_component.use_edma) {
          mca_btl_hlink_component.mpm_transport_hyplnk0 = (void *)mpm_transport_open("arm-remote-hyplnk-0", &ocfg);
        } else {
          mca_btl_hlink_component.mpm_transport_hyplnk0 = (void *)mpm_transport_open("arm-remote-hyplnk-0-no-dma", &ocfg);
        }
        if (!mca_btl_hlink_component.mpm_transport_hyplnk0) {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 open failed\n");
          mca_btl_hlink_component.use_hl0 = 0;
        } else {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 open successfull!!");
        }
      }
   }

   timpi_dbg_hlink(1);
   /*********************************************************************************************/
   opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK hyplnk0=%p hyplnk1=%p\n", 
               mca_btl_hlink_component.mpm_transport_hyplnk0, mca_btl_hlink_component.mpm_transport_hyplnk1);
      mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname = NULL;
      mca_btl_hlink_component.hl_win_rmt_rd_bridge          = NULL;
      mca_btl_hlink_component.hl_win_rmt_wr_bridge          = NULL;
      /*********************************************************************************************/
      /** HYPLNK0 ACCESS: one remote MSMC region and one remote DDR regions is mapped via hyplnk0 **/
      /*********************************************************************************************/
      if(mca_btl_hlink_component.use_hl0)
      {
        timpi_dbg_hlink_health(100);
        /** Map remote MSMC region visible via hyplnk0 to user space **/
        mca_btl_hlink_component.hl_win_msmc_rmt = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk0, 
                                                                      mca_btl_hlink_component.msmc_rmt_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
        if(!mca_btl_hlink_component.hl_win_msmc_rmt)
        {
          opal_output(0, "mmap_failed_hl_win_msmc_rmt (MSMC over hyplnk0)!");
          return NULL;
        }
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 MSMC mmap successfull!!");
        timpi_dbg_hlink_health(101);
        /** Map remote DDR region visible via hyplnk0 to user space **/
        mca_btl_hlink_component.hl_win_rmt = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk0,
                                                                 mca_btl_hlink_component.ddr_rmt_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
        if(!mca_btl_hlink_component.hl_win_rmt)
        {
          opal_output (0, "mmap_failed_hl_win_rmt (DDR over hyplnk0)!");
          return NULL;
        }
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 DDR mmap successfull!!");
        timpi_dbg_hlink_health(102);
        /* This is where I am going to read remote hostname visible over hyplnk0 */
        mca_btl_hlink_component.hl_win_rmt_hostname = (void *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt + BTL_HLINK_MSG_OFFSET - MAX_HOSTNAME_SIZE);

      timpi_dbg_hlink(10);
      if(mca_btl_hlink_component.use_hl0 * mca_btl_hlink_component.use_hl1)
      {
        /*** DIAGONAL connection ***/
        /** Map remote MSMC region visible via hyplnk0 to user space **/
        timpi_dbg_hlink_health(200);
        if(mca_btl_hlink_component.node_id == '1')
        {  /** Node '1' does writting over Hyperlink0 **/
           mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk0,
                                                                        mca_btl_hlink_component.msmc_rmt_bridge_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
           if(!mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge)
           {
             opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK mmap_failed_hl_win_msmc_rmt_bridge (MSMC over hyplnk0)!");
             return NULL;
           }
memset(mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge, 0, BTL_HLINK_MSMC_MMAP_SIZE);
btl_hlink_atomic_smb();
           opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 MSMC bridge wr mmap successfull!!");
           timpi_dbg_hlink_health(201);
          /** Map remote DDR region visible via hyplnk0 to user space **/
          mca_btl_hlink_component.hl_win_rmt_wr_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk0,
                                                                 mca_btl_hlink_component.ddr_rmt_bridge_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
          if(!mca_btl_hlink_component.hl_win_rmt_wr_bridge)
          {
            opal_output (0, "mmap_failed_hl_win_rmt_wr (DDR over hyplnk0)!");
            return NULL;
          }
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 DDR mmap wr successfull!!");
          timpi_dbg_hlink_health(202);
          /* This is where I am indicating identity accross diagonal connection */
          /* For diagonal connections, writting is always done over hyperlink0  */
          ptr_hostname = 
            (void *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge + BTL_HLINK_MSG_OFFSET - MAX_HOSTNAME_SIZE);
          memset (stage_hostname, 0, MAX_HOSTNAME_SIZE);
          strcpy (stage_hostname, host_name);
          *((unsigned int *)(stage_hostname + MAX_HOSTNAME_SIZE - sizeof(unsigned int))) = SYNC_TOKEN;
          btl_hlink_atomic_smb();
          memcpy (ptr_hostname, stage_hostname, MAX_HOSTNAME_SIZE);
          btl_hlink_atomic_smb();
        } else { /* Nodes '2','3' and '4' do reading over HL0 */
           timpi_dbg_hlink_health(300);
           mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk0,
                                                                        mca_btl_hlink_component.msmc_rmt_bridge_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
           if(!mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge)
           {
             opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK mmap_failed_hl_win_msmc_rmt_rd_bridge (MSMC over hyplnk0)!");
             return NULL;
           }
           opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 MSMC bridge rd mmap successfull!!");
           timpi_dbg_hlink_health(301);
          /** Map remote DDR region visible via hyplnk0 to user space **/
          mca_btl_hlink_component.hl_win_rmt_rd_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk0,
                                                                 mca_btl_hlink_component.ddr_rmt_bridge_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
          if(!mca_btl_hlink_component.hl_win_rmt_rd_bridge)
          {
            opal_output (0, "mmap_failed_hl_win_rmt_rd_bridge (DDR over hyplnk0)!");
            return NULL;
          }
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 DDR mmap successfull!!");
          timpi_dbg_hlink_health(302);
          mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname =
            (void *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge + BTL_HLINK_MSG_OFFSET - MAX_HOSTNAME_SIZE);
        }
      }
      }
      /*********************************************************************************************/
      /** HYPLNK1 ACCESS: one remote MSMC region and one remote DDR regions is mapped via hyplnk1 **/
      /*********************************************************************************************/
      if(mca_btl_hlink_component.use_hl1)
      {
        timpi_dbg_hlink_health(400);
        /** Map remote MSMC region visible via hyplnk1 to user space **/
        mca_btl_hlink_component.hl_win_msmc_rmt1 = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk1, 
                                                                      mca_btl_hlink_component.msmc_rmt1_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
        if(!mca_btl_hlink_component.hl_win_msmc_rmt1)
        {
          opal_output(0, "mmap_failed_hl_win_msmc_rmt (MSMC over hyplnk1)!");
          return NULL;
        }
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 MSMC mmap successfull!!");
        timpi_dbg_hlink_health(401);
        /** Map remote DDR region visible via hyplnk1 to user space **/
        mca_btl_hlink_component.hl_win_rmt1 = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk1,
                                                                  mca_btl_hlink_component.ddr_rmt1_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
        if(!mca_btl_hlink_component.hl_win_rmt1)
        {
          opal_output(0, "mmap_failed_hl_win_rmt (DDR over hyplnk1)!\n");
          return NULL;
        }
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 DDR mmap successfull!!");
        timpi_dbg_hlink_health(402);
        /* This is where I am going to read remote hostname visible over hyplnk1 */
        mca_btl_hlink_component.hl_win_rmt1_hostname = (void *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt1 + BTL_HLINK_MSG_OFFSET - MAX_HOSTNAME_SIZE);

        timpi_dbg_hlink(11);

      if(mca_btl_hlink_component.use_hl0 * mca_btl_hlink_component.use_hl1)
      {
        /*** DIAGONAL connection over Hyperlink1 ***/
        if(mca_btl_hlink_component.node_id == '1')
        {  /** Node '1' does reading over Hyperlink1 **/
           mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk1,
                                                                        mca_btl_hlink_component.msmc_rmt_bridge_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
           if(!mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge)
           {
             opal_output(0, "mmap_failed_hl_win_msmc_rmt_rd_bridge (MSMC over hyplnk1)!");
             return NULL;
           }
           opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 MSMC bridge rd mmap successfull!!");
          /** Map remote DDR region visible via hyplnk1 to user space **/
          mca_btl_hlink_component.hl_win_rmt_rd_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk1,
                                                                 mca_btl_hlink_component.ddr_rmt_bridge_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
          if(!mca_btl_hlink_component.hl_win_rmt_rd_bridge)
          {
            opal_output (0, "mmap_failed_hl_win_rmt_rd (DDR over hyplnk1)!");
            return NULL;
          }
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 DDR mmap rd successfull!!");
          /* This is where I am indicating identity accross diagonal connection */
          /* For diagonal connections, writting is always done over hyperlink0  */
          mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname =
            (void *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge + BTL_HLINK_MSG_OFFSET - MAX_HOSTNAME_SIZE);
        } else { /* Nodes '2','3' and '4' do writting over HL1 */
           mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk1,
                                                                        mca_btl_hlink_component.msmc_rmt_bridge_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
           if(!mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge)
           {
             opal_output(0, "mmap_failed_hl_win_msmc_rmt_wr_bridge (MSMC over hyplnk1)!");
             return NULL;
           }
memset(mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge, 0, BTL_HLINK_MSMC_MMAP_SIZE);
btl_hlink_atomic_smb();
           opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 MSMC bridge wr mmap successfull!!");
          /** Map remote DDR region visible via hyplnk0 to user space **/
          mca_btl_hlink_component.hl_win_rmt_wr_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk1,
                                                                 mca_btl_hlink_component.ddr_rmt_bridge_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
          if(!mca_btl_hlink_component.hl_win_rmt_wr_bridge)
          {
            opal_output (0, "mmap_failed_hl_win_rmt_wr_bridge (DDR over hyplnk1)!");
            return NULL;
          }
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 DDR mmap wr successfull!!");
          ptr_hostname = (void *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge + BTL_HLINK_MSG_OFFSET - MAX_HOSTNAME_SIZE);
          memset (stage_hostname, 0, MAX_HOSTNAME_SIZE);
          strcpy (stage_hostname, host_name);
          *((unsigned int *)(stage_hostname + MAX_HOSTNAME_SIZE - sizeof(unsigned int))) = SYNC_TOKEN;
          btl_hlink_atomic_smb();
          memcpy (ptr_hostname, stage_hostname, MAX_HOSTNAME_SIZE);
          btl_hlink_atomic_smb();
        }
      }
      }
/***SYNC NODES***/
//timpi_dbg_serdes_dump(700);
    timpi_dbg_hlink(12);
/***SYNC NODES***/

    if(!mca_btl_hlink_component.use_hl0 && !mca_btl_hlink_component.use_hl1)
    { /* If both interfaces are turned off, do not use hlink */
       opal_output(0, "HLINK turned off !!!");
       return NULL;
    }

    /* Set signal handlers to catch kill signals so we can properly clean up after ourselves. */
    //opal_event_set(opal_event_base, &btl_hlink_term_handler2, SIGINT, OPAL_EV_SIGNAL|OPAL_EV_PERSIST, btl_hlink_shutdown_callback, &btl_hlink_term_handler2);
    //opal_event_add(&btl_hlink_term_handler2, NULL);
   timpi_dbg_hlink(2);

   timpi_dbg_hlink_end();

    opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK BTL HLINK turned ON (hl0=%d hl1=%d)!!!", 
      mca_btl_hlink_component.use_hl0, mca_btl_hlink_component.use_hl1);
    return btls;
}

/*******************************************************************************************/
void mca_btl_hlink_find_fifo_range (int endpoint_count, int my_rank, int *j_start, int *j_end)
{
    *j_start = *j_end = 0;
return;
    if(endpoint_count == 4)
    {
      if((my_rank % 2) == 0)
      {
        *j_start = 0; *j_end = 1;
      } else {
        *j_start = 2; *j_end = 3;
      }
    } 
}
/*******************************************************************************************/
void mca_btl_hlink_one_fragment_received (mca_btl_hlink_hdr_t *hdr)
{
    mca_btl_hlink_frag_t Frag;
    mca_btl_active_message_callback_t* reg;
    size_t hdr_len;
    unsigned int token_index, *check_magic_token;
    int hdr_invalidate = (sizeof(mca_btl_hlink_hdr_t) + 127) & 0x00ffff80;
#ifdef BTL_HLINK_VERBOSE
    struct timespec tp_start, tp_end;
clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
    for (;;) {
    /* recv upcall */
#ifdef BTL_HLINK_USE_CACHED
      CMEM_cacheInv(hdr, hdr_invalidate);
#endif
      if((hdr->len < 0) || (hdr->len > 128*1024)) {
        opal_output_verbose(0, mca_btl_base_output, "Bad header:%08x", hdr->len);
      } else break;
    }
    memset (&Frag, 0, sizeof(mca_btl_hlink_frag_t));
    reg = mca_btl_base_active_message_trigger + hdr->tag;
    Frag.base.des_dst = &Frag.segment;
    Frag.segment.seg_addr.pval = ((char *)hdr) + sizeof(mca_btl_hlink_hdr_t);
    memcpy (&Frag.segment.seg_len, &hdr->len, sizeof(size_t));
    hdr_len = hdr->len;
    Frag.base.des_dst_cnt = 1;
    opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG:HLINK one_fragment before token search");
#if 1
    {
      int inv_cnt;
      token_index = btl_hlink_alignCACHELINE(sizeof(mca_btl_hlink_hdr_t) + hdr_len);
      check_magic_token = (unsigned int *)((char *)hdr + token_index); 
      for (inv_cnt = 0; ; inv_cnt ++)
      { 
#ifdef BTL_HLINK_USE_CACHED
         CMEM_cacheInv((void *)((char *)hdr + token_index), 256);
#endif
         if((check_magic_token[0] == 0xbabebeef) && (check_magic_token[1] == 0xdedababa)) break;
         if((check_magic_token[32] == 0xbabebeef) && (check_magic_token[33] == 0xdedababa)) 
         {
           check_magic_token += 32;
           break;
         }
         if((inv_cnt > 0) && (inv_cnt < 4))
         {
           opal_output_verbose(0, mca_btl_base_output, "hdr %p size_t %d BTL_HLINK one_fragement magic token:%08x %08x, seg_len=%d token_index=%d", 
             hdr, sizeof(size_t), check_magic_token[0], check_magic_token[1], hdr_len, token_index);
           opal_output_verbose(0, mca_btl_base_output, "[%08x %08x %08x %08x]", ((unsigned int *)hdr)[0], ((unsigned int *)hdr)[1],
             ((unsigned int *)hdr)[2], ((unsigned int *)hdr)[3]);
         }
       }
    }
#endif
#ifdef BTL_HLINK_USE_CACHED
    CMEM_cacheInv(hdr, token_index);
#endif

#ifdef BTL_HLINK_VERBOSE
clock_gettime(CLOCK_MONOTONIC, &tp_end);
opal_output_verbose(10, mca_btl_base_output, "BTL_HLINK cacheInv(two times): %lf us for %d bytes", 
                    1000.0 * clock_diff(&tp_start, &tp_end), Frag.segment.seg_len);
#endif
    opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG:HLINK one_fragment before cbfunc");
    reg->cbfunc(&mca_btl_hlink.super, hdr->tag, &(Frag.base), reg->cbdata);
    opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG:HLINK one_fragment FRAG RCV end");
#ifdef BTL_HLINK_USE_CACHED
    check_magic_token[0] = check_magic_token[1] = 0;
    CMEM_cacheWb((char *)check_magic_token, 128);
#endif
}
/*******************************************************************************************/
void mca_btl_hlink_one_fragment_received_bridge (mca_btl_hlink_hdr_t *hdr)
{
    mca_btl_hlink_frag_t Frag;
    mca_btl_active_message_callback_t* reg;
    mca_btl_hlink_hdr_t loc_hdr;
    char *stage_buffer = (char *)(mca_btl_hlink_component.hl_win_edma_transfer + 0x00100000);
    //Copy buffer from remote DDR to local DDR 
    loc_hdr = *hdr;
    memcpy (stage_buffer, hdr, loc_hdr.len + sizeof(mca_btl_hlink_hdr_t) + 16);
    reg = mca_btl_base_active_message_trigger + loc_hdr.tag;
    Frag.base.des_dst = &Frag.segment;
    Frag.segment.seg_addr.pval = stage_buffer + sizeof(mca_btl_hlink_hdr_t);
    Frag.segment.seg_len = loc_hdr.len;
    Frag.base.des_dst_cnt = 1;
    //btl_hlink_atomic_smb();
    reg->cbfunc(&mca_btl_hlink.super, hdr->tag, &(Frag.base), reg->cbdata);
    opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG:HLINK component_progress FRAG RCV end");
}
/*******************************************************************************************/
static int mca_btl_hlink_component_progress (void)
{
  mca_btl_hlink_hdr_t *hdr, *hdr_offset;
  int peer_rank, rc = 0, nevents = 0;
  int j, j_start, j_end;
  int ack_frag = 0;
  char *ptr;
  mca_btl_hlink_find_fifo_range(mca_btl_hlink_component.endpoint_count, mca_btl_hlink_component.my_rank, &j_start, &j_end);

timpi_dbg_hlink_health(500);

  if(mca_btl_hlink_component.use_hl0)
  {
  //=====================
  // Hyperlink0 interface
  //=====================
    for(j = j_start; j <= j_end; j ++)
    {
timpi_dbg_hlink_health(510);
        hdr_offset = (mca_btl_hlink_hdr_t *)hlink_fifo_read_begin(mca_btl_hlink_component.fifo[0][j]);
        /* change the address from address relative to the shared
         * memory address, to a true virtual address */
        hdr = RELATIVE2VIRTUAL_0(hdr_offset, j);
        if(hdr == NULL) continue;
opal_output_verbose(10, mca_btl_base_output, "BTL_HLINK RCV hdr_HL0=%p", hdr_offset);
timpi_dbg_hlink_health(511);
        mca_btl_hlink_one_fragment_received (hdr);
timpi_dbg_hlink_health(512);
        hlink_fifo_read_end(mca_btl_hlink_component.fifo[0][j]);
        nevents ++;
#ifdef BTL_HLINK_VERBOSE
   opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG component_progress reading Hyperlink0");
#endif
    }
  }

  if(mca_btl_hlink_component.use_hl1)
  {
    //=====================
    // Hyperlink1 interface
    //=====================
    for(j = j_start; j <= j_end; j ++)
    {
timpi_dbg_hlink_health(520);
        hdr_offset = (mca_btl_hlink_hdr_t *)hlink_fifo_read_begin(mca_btl_hlink_component.fifo1[0][j]);
        /* change the address from address relative to the shared
         * memory address, to a true virtual address */
        hdr = RELATIVE2VIRTUAL_1(hdr_offset, j);
        if(hdr == NULL) continue;
opal_output_verbose(10, mca_btl_base_output, "BTL_HLINK RCV hdr_HL1=%p", hdr_offset);
timpi_dbg_hlink_health(521);
        mca_btl_hlink_one_fragment_received (hdr);
timpi_dbg_hlink_health(522);
        hlink_fifo_read_end(mca_btl_hlink_component.fifo1[0][j]);
        nevents ++;
#ifdef BTL_HLINK_VERBOSE
    opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG component_progress reading Hyperlink1");
#endif
    }

    if(mca_btl_hlink_component.endpoint_count > 2 && mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge)
    {
      for(j = j_start; j <= j_end; j ++)
      {
timpi_dbg_hlink_health(530);
        hdr_offset = (mca_btl_hlink_hdr_t *)hlink_fifo_read_begin(mca_btl_hlink_component.fifo_bridge[0][j]);
        hdr = RELATIVE2VIRTUAL_BRIDGE(hdr_offset, j);
        if(hdr == NULL) continue;
opal_output_verbose(10, mca_btl_base_output, "BTL_HLINK RCV hdr_HL_bridge=%p", hdr_offset);
timpi_dbg_hlink_health(531);
        mca_btl_hlink_one_fragment_received_bridge (hdr);
timpi_dbg_hlink_health(532);
        hlink_fifo_read_end(mca_btl_hlink_component.fifo_bridge[0][j]);
        nevents ++;
#ifdef BTL_HLINK_VERBOSE
      opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG component_progress reading diagonal node");
#endif
      }  
    }
  }
timpi_dbg_hlink_health(502);
  return 1;
}
/*** nothing past this point ***/
