#ifndef MCA_BTL_HLINK_FIFO_H
#define MCA_BTL_HLINK_FIFO_H

/* We can't use opal_cache_line_size here because we need a
   compile-time constant for padding the struct.  We can't really have
   a compile-time constant that is portable, either (e.g., compile on
   one machine and run on another).  So just use a big enough cache
   line that should hopefully be good in most places. */
#define HLINK_CACHE_LINE_PAD 128


struct hlink_fifo_t {
    unsigned int id;
    unsigned int qsize;
    unsigned int head_cnt;
    //char pad0[HLINK_CACHE_LINE_PAD - 2 * sizeof(unsigned int)];
    unsigned int tail_cnt;
    //char pad1[HLINK_CACHE_LINE_PAD - 1 * sizeof(unsigned int)];
    /* This index is used by the head holding the head lock. */
    void *msg_offset[1];
};

typedef struct hlink_fifo_t hlink_fifo_t;
/***
 * One or more FIFO components may be a pointer that must be
 * accessed by multiple processes.  Since the shared region may
 * be mmapped differently into each process's address space,
 * these pointers will be relative to some base address.  Here,
 * we define macros to translate between relative addresses and
 * virtual addresses.
 */
#define RELATIVE2VIRTUAL_0(OFFSET, n) (OFFSET == -1 ? NULL : (long)(OFFSET) + (long)mca_btl_hlink_component.hl_win_loc_msgbase[n])
#define RELATIVE2VIRTUAL_1(OFFSET, n) (OFFSET == -1 ? NULL : (long)(OFFSET) + (long)mca_btl_hlink_component.hl_win_loc1_msgbase[n])
#define RELATIVE2VIRTUAL_BRIDGE(OFFSET, n) (OFFSET == -1 ? NULL : (long)(OFFSET) + (long)mca_btl_hlink_component.hl_win_rmt_rd_bridge_msgbase[n])

/*** Memory sync barrier ***/
static inline void btl_hlink_atomic_smb(void) {
  __asm__ __volatile__ ("dsb" : : : "memory");
}

static inline int hlink_fifo_getsize(int fifo_size)
{
    return sizeof(void *) * (fifo_size - 1) + sizeof(hlink_fifo_t);
}

static inline int hlink_fifo_init(void *base_address, int fifo_size)
{
    int qsize;
    hlink_fifo_t *fifo;
    
    qsize = hlink_fifo_getsize(fifo_size);
    memset (base_address, 0, qsize);
    fifo = (hlink_fifo_t *)base_address;
    fifo->qsize = fifo_size;
    fifo->id    = 0xbebadeba;
    return OMPI_SUCCESS;
}

static inline int hlink_fifo_write(hlink_fifo_t *fifo, int value)
{
    int fifo_qsize;
    int head_cnt;
    int tail_cnt;
    int fifo_entries;

    fifo_qsize = fifo->qsize;
    head_cnt = fifo->head_cnt;
    tail_cnt = fifo->tail_cnt;
    fifo_entries  = head_cnt - tail_cnt;
    /* if there is no free slot to write, report exhausted resource */
    if(fifo_entries >= fifo_qsize) {
      return OMPI_ERR_OUT_OF_RESOURCE;
    }
    /* otherwise, write to the slot and advance the head index */
    fifo->msg_offset[head_cnt % fifo_qsize] = value;
    head_cnt ++;
    fifo->head_cnt = head_cnt;
    btl_hlink_atomic_smb();

    return OMPI_SUCCESS;
}


static inline int hlink_fifo_acquire(hlink_fifo_t *fifo, int value)
{
    int fifo_qsize;
    int head_cnt;
    int tail_cnt;
    int fifo_entries;

    fifo_qsize = fifo->qsize;
    head_cnt = fifo->head_cnt;
    tail_cnt = fifo->tail_cnt;
    fifo_entries  = head_cnt - tail_cnt;
    /* if there is no free slot to write, report exhausted resource */
    if(fifo_entries >= fifo_qsize) {
      return -1;
    }
    /* otherwise, write to the slot and advance the head index */
    fifo->msg_offset[head_cnt % fifo_qsize] = value;
    btl_hlink_atomic_smb();

    return head_cnt;
}
static inline void *hlink_fifo_read(hlink_fifo_t *fifo)
{
    int fifo_qsize;
    int head_cnt;
    int tail_cnt;
    int fifo_entries;
    void *value = -1;

    fifo_qsize = fifo->qsize;
    head_cnt = fifo->head_cnt;
    tail_cnt = fifo->tail_cnt;
    fifo_entries  = head_cnt - tail_cnt;
    /* if there is no free slot to write, report exhausted resource */
    if(fifo_entries > 0) {
      value = fifo->msg_offset[tail_cnt % fifo_qsize];
      tail_cnt ++;
      fifo->tail_cnt = tail_cnt;
      btl_hlink_atomic_smb();
    }
    return value;
}

static inline void *hlink_fifo_read_begin(hlink_fifo_t *fifo)
{
    int fifo_qsize;
    int head_cnt;
    int tail_cnt;
    int fifo_entries;
    void *value = -1;

    fifo_qsize = fifo->qsize;
    head_cnt = fifo->head_cnt;
    tail_cnt = fifo->tail_cnt;
    fifo_entries  = head_cnt - tail_cnt;
    /* if there is no free slot to write, report exhausted resource */
    if(fifo_entries > 0) {
      value = fifo->msg_offset[tail_cnt % fifo_qsize];
    }
    return value;
}
static inline void hlink_fifo_read_end(hlink_fifo_t *fifo)
{
   int tail_cnt = fifo->tail_cnt;
   tail_cnt ++;
   fifo->tail_cnt = tail_cnt;
   btl_hlink_atomic_smb();
}
#endif
