#include "mpi.h" 
#include <stdlib.h> 
#include <stdio.h> 
#include <string.h>
#include <math.h> 

/* #define VERBOSE */
#define ITERRATION_CNT 20
#define MAX_NBODY 1000 
#define MAX_TOTAL_NBODY 4000 
/* Detect collision if bodies are closer than MIN_DISTANCE */
#define MIN_DISTANCE   0.001
/* Time step: fixed */
#define TIME_STEP 0.1
/* Range of random locations along x, y and z axe: -XSPAN...+XSPAN */
#define XSPAN 10.0
#define YSPAN 10.0
#define ZSPAN 0.0
/* Range of random velocities along x, y and z axe: -XVSPAN ... +XVSPAN */
#define XVSPAN 5.0
#define YVSPAN 5.0
#define ZVSPAN 0.0
/* Min and max mass of body */
#define MIN_MASS 1.0
#define MAX_MASS 3.0

#if 1
double drand48(void) {
  return (double)rand() / 32768.0;
}
#endif
/**************************************************************************/
typedef struct { 
    double x, y, z;    /* 3D Location */
    double xold, yold, zold;
    double m;  /* mass */ 
} Body_t; 

typedef struct { 
    double fx, fy, fz;
} Body_Forces_t;
/*********************************/
/* Initialize body positions */
/*********************************/ 
void setup_nbody( Body_t nbody[], int npart ) 
{ 
    int i;
    double vx, vy, vz;
	
    for (i=0; i<npart; i++) 
    { 
      nbody[i].x  = -XSPAN + 2 * XSPAN * drand48(); 
      nbody[i].y  = -YSPAN + 2 * YSPAN * drand48(); 
      nbody[i].z  = -ZSPAN + 2 * ZSPAN * drand48(); 
      nbody[i].m  = MIN_MASS + (MAX_MASS - MIN_MASS) * drand48(); /* mass is in 1.0-2.0 range */ 	  
      /* random velocities */
      vx          = -XVSPAN + 2 * XVSPAN * drand48(); 
      vy          = -YVSPAN + 2 * YVSPAN * drand48(); 
      vz          = -ZVSPAN + 2 * ZVSPAN * drand48(); 
      nbody[i].xold = nbody[i].x  - vx * TIME_STEP;	  
      nbody[i].yold = nbody[i].y  - vy * TIME_STEP;	  
      nbody[i].zold = nbody[i].z  - vz * TIME_STEP;	  
    } 
}
void reset_nbody_forces (Body_Forces_t forces_loc[], int npart)
{
    int i; 
    for (i=0; i<npart; i++) 
    {
      forces_loc[i].fx = forces_loc[i].fy = forces_loc[i].fz = 0.0; 	
    }
} 
/************************************************/
/* Compute forces for list of nbody         */
/* This is a good candidate for DSP offload     */
/************************************************/
void compute_nbody_force ( int rank, Body_t nbody[], Body_Forces_t forces_loc[], 
                           Body_t remote_nbody[], int npart)
{						   
  double remote_m;
  double x, y, z;
  double rx, ry, rz, r;
  double fx, fy, fz; 
  int    i, j; 
 
  for (i = 0; i < npart; i ++) 
  { 
    x   = nbody[i].x; 
    y   = nbody[i].y; 
    z   = nbody[i].z; 

    fx = fy = fz = 0.0; 
    
    for (j = 0; j < npart; j++)
    { 
      /* Find distance between two nbody */
      rx = x - remote_nbody[j].x; 
      ry = y - remote_nbody[j].y; 
      rz = z - remote_nbody[j].z; 	  
      r  = rx * rx + ry * ry + rz * rz; 
      if(r == 0.0) continue; /* same body */

      remote_m = remote_nbody[j].m; 
      /* Treat collisions as anihilations: */
      if (r < MIN_DISTANCE) {
         printf ("\nCollision! Rank %d: [%d vs %d] distance %lf less than min(%lf).", 
		   rank, i, j, r, MIN_DISTANCE);
         printf ("\n\t\tNo support for elastic or non-elastic collision (anihilate!)");		   
         fflush(stdout);
	 fx = fy = fz = 0.0;
	 nbody[i].m   = 0.0;
	 forces_loc[i].fx = forces_loc[i].fy = forces_loc[i].fz = 0.0;
	 break;
      }
      /* compute & project forces */ 
      r  = r * sqrt(r); 
      fx -= remote_m * rx / r; /* split force into hor. component, and accumulate */
      fy -= remote_m * ry / r; /* split force into ver. component, and accumulate */
      fz -= remote_m * rz / r; /* split force into tranv. component, and accumulate */
    } 
	/* keep accumulating forces (contribution from peer nodes) */
    forces_loc[i].fx += fx;
    forces_loc[i].fy += fy; 
    forces_loc[i].fz += fz; 	
  } 
} 
/*****************************************************/
/* Update positions of all the nbody in the system   */
/*****************************************************/
double update_nbody_position( Body_t nbody[], Body_Forces_t forces_loc[], int npart ) 
{ 
  int i; 
  double next_x, next_y, next_z;  
  static double dt = TIME_STEP; 
 
  for (i= 0; i < npart; i++) 
  { 	
    /* Verlet integration */  	
    /* Velocities can be indirectly calculated as vx = (x - xold) / dt, etc.. */
    next_x = 
	  forces_loc[i].fx * dt * dt + 2 * nbody[i].x - nbody[i].xold;
    next_y = 
	  forces_loc[i].fy * dt * dt + 2 * nbody[i].y - nbody[i].yold;
    next_z = 
	  forces_loc[i].fz * dt * dt + 2 * nbody[i].z - nbody[i].zold;

    nbody[i].xold = nbody[i].x;
    nbody[i].yold = nbody[i].y;
    nbody[i].zold = nbody[i].z;

    nbody[i].x = next_x;
    nbody[i].y = next_y;
    nbody[i].z = next_z;
  }
  
  /* reset forces that were accumulated, prepare for next time step */
  reset_nbody_forces (forces_loc, npart);
  return dt;
}

/*****************************************************************************/
void main( int argc, char *argv[] ) 
{
    int rank, size, npart;
    int	i, j, cnt;
    double start_time, simulation_time;  /* Start of Computation and simulation time */
    Body_t all_nbody[MAX_TOTAL_NBODY];
    Body_t nbody[MAX_NBODY];   /* nBody on ALL nodes */ 
    Body_Forces_t forces_loc[MAX_NBODY];   /* Body Forces on local nodes */ 
    Body_t nbody_to_next[MAX_NBODY], nbody_from_previous[MAX_NBODY];  /* Pipeline buffers */
    MPI_Datatype bodytype;      /* derived type to be promoted to nbody */
    MPI_Status  statuses[2]; 
    MPI_Request request[2]; 
 
    MPI_Init( &argc, &argv ); 
    MPI_Comm_rank( MPI_COMM_WORLD, &rank ); 
    MPI_Comm_size( MPI_COMM_WORLD, &size ); 
 
    if (argc < 2) {  
	  fprintf( stderr, "Usage: %s n\n", argv[0] ); 
	  MPI_Abort( MPI_COMM_WORLD, 1 ); 
    }
	/* nBody per node */
    npart = atoi(argv[1]) / size; 

    if (npart > MAX_NBODY) { 
	  fprintf( stderr, "Maximum exceeded. Simulating maximum of %d nbody per process\n", MAX_NBODY); 
	  MPI_Abort( MPI_COMM_WORLD, 2 ); 
    } 
    if ((npart * size) > MAX_TOTAL_NBODY) { 
	  fprintf( stderr, "Maximum exceeded. Simulating maximum total of %d nbody (all processes)\n", MAX_TOTAL_NBODY); 
	  MPI_Abort( MPI_COMM_WORLD, 2 ); 
    } 
    MPI_Type_contiguous( 7, MPI_DOUBLE, &bodytype ); 
    MPI_Type_commit( &bodytype ); 
 
    /* Setup the initial state - every node creates collection 
       of its own nbody */ 
    if(rank == 0) {
      /* Only master node generates initial positions */
      setup_nbody ( all_nbody, npart * size); 
    }
    MPI_Scatter(all_nbody, npart, bodytype, 
	            nbody, npart, bodytype, 
    		    0, MPI_COMM_WORLD);
    reset_nbody_forces (forces_loc, npart);
    /* Capture start of simulation */
    start_time = MPI_Wtime(); 
    simulation_time = 0.0;
    for (cnt = 0; cnt < ITERRATION_CNT; cnt ++) 
    { 
       /* Load the initial nbody_to_next */ 
       memcpy( nbody_to_next, nbody, npart * sizeof(Body_t) ); 
	  
       for (i = 0; i < (size - 1); i ++) 
       { 
	  /* ...exchange data in round-robin fashion. 
	   * This is also non-blocking send&receive.
	   * It will happen in parallel with force computations. */
	   MPI_Isend( nbody_to_next, npart, bodytype, (rank + 1) % size, i,  
		           MPI_COMM_WORLD, &request[0] ); 
	   MPI_Irecv( nbody_from_previous, npart, bodytype, (rank + size - 1) % size, i,  
		           MPI_COMM_WORLD, &request[1] ); 
				   
	   /* Compute forces (2D only) */ 
	   compute_nbody_force( rank, nbody, forces_loc, nbody_to_next, npart );
	   /* wait for the above two non-blocking operations to finish */ 
	   MPI_Waitall( 2, request, statuses ); 
#ifdef VERBOSE
printf("\nRANK: %d %lf %lf %lf (mass=%lf) recv %lf %lf %lf (%lf)", rank, 
	   nbody_to_next[0].x, nbody_to_next[0].y, nbody_to_next[0].z, nbody_to_next[0].m,
		nbody_from_previous[0].x, nbody_from_previous[0].y, nbody_from_previous[0].z, nbody_from_previous[0].m);
#endif
	   memcpy( nbody_to_next, nbody_from_previous, npart * sizeof(Body_t) ); 
        } 
	/* Compute forces (2D only) */ 
	compute_nbody_force( rank, nbody, forces_loc, nbody_to_next, npart ); 
	/* Once we have the forces, we compute the changes in position */ 
	simulation_time += update_nbody_position ( nbody, forces_loc, npart );
    }
    if (rank == 0) { /* Only master core does the report: */
	printf( "\nSimulation of %lf seconds done in %lf seconds\n", simulation_time, MPI_Wtime() - start_time ); 
    } 
    MPI_Type_free (&bodytype);
    MPI_Finalize(); 
    return;
} 
/* nothing past this point */
